// vim:set tabstop=4 cindent:

// Author: Geert-Jan Giezeman <geert@cs.uu.nl>
// Copyright: Utrecht University, 2011

#include "fitting.hpp"
#define _USE_MATH_DEFINES 
#include <math.h>

using namespace UuIcsGeom;

namespace {

	Vector3d optimal_rotation_axis(std::vector<Vector3d> const &model,
			std::vector<Vector3d> const &object)
	{
		assert(model.size()==object.size());
		size_t const sz = model.size();
		Vector3d result(0,0,0);
		for (size_t i=0; i!= sz; ++i) {
			Vector3d const &m = model[i];
			Vector3d const &o = object[i];
			Vector3d diff(m);
			diff -= o;
			result.coords[0] +=
				o.coords[2]*diff.coords[1]-o.coords[1]*diff.coords[2];
			result.coords[1] +=
				o.coords[0]*diff.coords[2]-o.coords[2]*diff.coords[0];
			result.coords[2] +=
				o.coords[1]*diff.coords[0]-o.coords[0]*diff.coords[1];
		}
		double l = vector_length(result.coords);
		if (l <= 0.0) {
			result.coords[2] = 1.0;
		} else {
			result *= 1.0/l;
		}
		return result;
	}

	//! compute theta such that the sum of squared distances between 'model'
	//! and rotated object points (by theta around unit_axis) is minimal
	double optimal_theta(std::vector<Vector3d> model,
			std::vector<Vector3d> object, Vector3d const &unit_axis)
	{
		size_t sz=model.size();
		assert(object.size()==sz);
		// take an orthogonal basis where the unit_axis is the first vector and two arbitrary axes to complete it
		Matrix3d tfm1;
		Vector3d axis2, axis3;
		{
			double x=unit_axis.coords[0];
			double y=unit_axis.coords[1];
			double z=unit_axis.coords[2];
			double x_2=x*x, y_2=y*y, z_2=z*z;
			if (x_2<=y_2 && x_2<=z_2) {
				double il = 1.0/sqrt(y_2+z_2);
				axis2.set(0.0, -il*z, il*y);
				axis3.set(il*(y_2+z_2), -il*x*y, -il*z*x);
			} else if (y_2<=x_2 && y_2<=z_2) {
				double il = 1.0/sqrt(x_2+z_2);
				axis2.set(il*z, 0.0, -il*x);
				axis3.set(-il*x*y, il*(x_2+z_2),-il*z*y);
			} else {
				double il = 1.0/sqrt(x_2+y_2);
				axis2.set(-il*y, il*x, 0.0);
				axis3.set(-il*x*z, -il*z*y, il*(x_2+y_2));
			}
			tfm1.set(x,y,z,axis2.coords[0],
					axis2.coords[1],axis2.coords[2],
					axis3.coords[0],axis3.coords[1],axis3.coords[2]);
		}
		// represent the model and object in this basis
		for (size_t i=0; i!=sz; ++i) {
			model[i] = tfm1*model[i];
			object[i] = tfm1*object[i];
		}
		// Because in this basis the position of the object as function of theta
		// is (object[i].x, object[i].y*cos(theta)-object[i].z*sin(theta),
		// object[i].z*cos(theta)+object[i].y*sin(theta)) 
		// the distance as a fuction of theta can be written as 
		// a*sin(theta) + b*cos(theta) + c;
		double cosfac=0.0, sinfac=0.0;
		for (size_t i=0; i!=sz; ++i) {
			DTriple const &m=model[i].coords;
			DTriple const &o=object[i].coords;
			cosfac += -2*(m[1]*o[1] + m[2]*o[2]);
			sinfac += 2*(m[1]*o[2]-m[2]*o[1]);
		}
		// Find a value of theta that minimises the function.
		double theta = atan2(-sinfac,-cosfac);
		return theta;
	}

	Vector3d find_center(std::vector<Point3d> const &object)
	{
		Vector3d center;
		if (object.size()>0) {
			center = object[0];
			for (int i=1; i!= object.size(); ++i)
				center += object[i];
			center *= (1.0/object.size());
		}
		return center;
	}

	Vector3d find_center(std::vector<Vector3d> const &object)
	{
		Vector3d center;
		if (object.size()>0) {
			center = object[0];
			for (int i=1; i!= object.size(); ++i)
				center += object[i];
			center *= (1.0/object.size());
		}
		return center;
	}
}

FittingResult fit(std::vector<Vector3d> const &model,
		std::vector<Point3d> const &object)
{
    assert(model.size() == object.size());
    assert(model.size()>0);
    FittingResult result;
    // Compute the center of the object
    Vector3d center(object[0].coords);
    for (int i=1; i!= object.size(); ++i)
        center += object[i];
    center *= (1.0/object.size());
    // Translate the object so that the center lies in the origin.
    result.tr = -center;
    std::vector<Vector3d> object_vecs(object.size(), result.tr);
    for (int i=0; i!= object.size(); ++i) {
        object_vecs[i] += object[i].coords;
    }
    result.transformed_object = object_vecs;
    // Find the optimal rotation
    int const MaxIterations = 10;
    double const CutoffAngle = 0.01;
    result.rot = Matrix3d::identity();
    // TODO: look at numerical stability. 
    // Perhaps use quaternions to compose rotations
    for (int iteration=0; iteration !=MaxIterations; ++iteration) {
        Vector3d unit_axis = optimal_rotation_axis(
				model, result.transformed_object);
        double theta = optimal_theta(
				model, result.transformed_object, unit_axis);
        Matrix3d rot = rotation_matrix_from_unit_axis(unit_axis, theta);
        result.rot = rot*result.rot;
        for (int i=0; i!= object.size(); ++i)
            result.transformed_object[i] = result.rot*object_vecs[i];
        result.d = 0.0;
        for (int i=0; i!= object.size(); ++i) {
            Vector3d diff = model[i]-result.transformed_object[i];
            result.d += inprod(diff, diff);
        }
        if (abs(theta) <CutoffAngle)
            break;
    }
    return result;
}

std::vector<Vector3d>  make_model(
		std::vector< std::vector<Point3d> > const &instances)
{
    size_t const objectcount = instances.size();
    assert(instances.size()>0);
    size_t const pointcount = instances[0].size();
    for (size_t i=0; i<objectcount; ++i) {
        assert(instances[i].size()==pointcount);
    }
    // take centered first object as initial model
    std::vector<Vector3d> model(pointcount);
    for (size_t i=0; i<pointcount; ++i)
        model[i] = instances[0][i];
    Vector3d center = find_center(model);
    for (size_t i=0; i<pointcount; ++i)
        model[i] -= center;
    // iterate to find better models
    int const IterationCount = 3;
    for (int iteration=0; iteration !=IterationCount; ++iteration) {
        // fit other instances to the current model
        std::vector<Vector3d> diff(pointcount, Vector3d(0,0,0));
        for (size_t i=0; i<objectcount; ++i) {
            FittingResult fr=fit(model,instances[i]);
            for (size_t p=0; p!= pointcount; ++p) {
                diff[p] += (1.0/objectcount)*(fr.transformed_object[p]-model[p]);
            }
        }
        // adapt model
        for (size_t p=0; p!= pointcount; ++p) {
            model[p] += diff[p];
        }
    }
    return model;
}

void check_fitting()
{
    using std::vector;
    size_t const sz = 4;
    vector<Vector3d> model(sz);
    model[0] = Vector3d(-1, 6, 2);
    model[1] = Vector3d(-2, -2, 1);
    model[2] = Vector3d(5, -1, 3);
    model[3] = Vector3d(-2, -3, -6);
    vector<Point3d> object(4);
    for (size_t i=0; i<sz; ++i) {
        object[i] = Point3d(model[i]);
    }
    FittingResult fr = fit(model, object);
    object[0].coords[1] += 0.1;
    object[3].coords[1] -= 0.1;
    fr = fit(model, object);
    Matrix3d tfm=rotation_matrix_from_axis(Vector3d(1,2,3), 1.9);
    Vector3d trl(-3.6, 1.0, 2.5);
    for (size_t i=0; i<sz; ++i) {
        object[i] = Point3d(tfm*object[i]+trl);
    }
    fr = fit(model, object);
}
