// vim:set tabstop=4 cindent:

// Author: Geert-Jan Giezeman <geert@cs.uu.nl>
// Copyright: Utrecht University, 2011

#include "geometry.hpp"

#define _USE_MATH_DEFINES 
#include <math.h>

namespace Geom=UuIcsGeom;

double Geom::inprod(Geom::DTriple const &v1, Geom::DTriple const &v2)
{
	double result=0.0;
	for (int i=0; i!=3; ++i)
		result += v1[i]*v2[i];
	return result;
}

Geom::Vector3d Geom::cross(Geom::DTriple const &vec0, Geom::DTriple const &vec1)
{
	Vector3d vec2;
	vec2.coords[0] = vec0[1]*vec1[2]-vec0[2]*vec1[1];
	vec2.coords[1] = vec0[2]*vec1[0]-vec0[0]*vec1[2];
	vec2.coords[2] = vec0[0]*vec1[1]-vec0[1]*vec1[0];
	return vec2;
}

double Geom::vector_length(Geom::DTriple const &vec)
{
	return sqrt(inprod(vec, vec));
}

Geom::Vector3d Geom::diffvec(Geom::DTriple const &pt1, Geom::DTriple const &pt2)
{
	Vector3d result;
	for (int i=0; i!=3; ++i)
		result.coords[i] = pt2[i]-pt1[i];
	return result;
}

namespace {
Geom::Vector3d a_perpendicular(Geom::DTriple const &vec1)
{
	Geom::Vector3d result;
	Geom::DTriple &rc = result.coords;
	double a = fabs(vec1[0]);
	double b = fabs(vec1[1]);
	double c = fabs(vec1[2]);
	if (a>=b) {
		if (a>=c) {
			rc[0] = vec1[1];
			rc[1] = -vec1[0];
			rc[2] = 0.0;
		} else {
			rc[0] = -vec1[2];
			rc[1] = 0.0;
			rc[2] = vec1[0];
		}
	} else {
		if (b>=c) {
			rc[0] = vec1[1];
			rc[1] = -vec1[0];
			rc[2] = 0.0;
		} else {
			rc[0] = -vec1[2];
			rc[1] = 0.0;
			rc[2] = vec1[0];
		}
	}
	return result;
}
}

Geom::Matrix3d const &Geom::Matrix3d::identity()
{
	static Matrix3d v(1,0,0, 0,1,0, 0,0,1);
	return v;
}


Geom::Vector3d Geom::mult(Geom::Matrix3d const &m, Geom::DTriple const &v)
{
	Vector3d result;
	int midx=0;
	for (int i=0; i!=3; ++i) {
		double sum = 0.0;
		for (int j=0; j!=3; ++j, ++midx) {
			sum += m.coords[midx]*v[j];
		}
		result.coords[i] = sum;
	}
	return result;
}

Geom::Matrix3d Geom::matrixmult(Geom::DNonet const &m1, Geom::DNonet const &m2)
{
	Matrix3d result;
	for (int i=0; i!=3; ++i) {
		for (int k=0; k!=3; ++k) {
			double sum = 0.0;
			for (int j=0; j!=3; ++j) {
				sum += m1[3*i+j]*m2[3*j+k];
			}
			result.set_value_raw(i,k, sum);
		}
	}
	return result;
}

Geom::Matrix3d Geom::rotation_matrix_from_unit_axis(
				Geom::DTriple const &r, double rad)
{
	Matrix3d result;
	double c = cos(rad);
	double s = sin(rad);
	double C = 1.0-c;
	double x = r[0];
	double y = r[1];
	double z = r[2];
	return Matrix3d(x*x*C+c, x*y*C-z*s, x*z*C+y*s, y*x*C+z*s,
            y*y*C+c, y*z*C-x*s,  z*x*C-y*s, z*y*C+x*s, z*z*C+c);
}

Geom::Matrix3d Geom::rotation_matrix_from_axis(Geom::Vector3d v, double rad)
{
	DTriple &r = v.coords;
	double length = sqrt(r[0]*r[0]+r[1]*r[1]+r[2]*r[2]);
	if (length == 0) {
		r[0]= 1.0;
		rad = 0.0;
	} else {
		r[0] = r[0]/length;
		r[1] = r[1]/length;
		r[2] = r[2]/length;
	}
	return rotation_matrix_from_unit_axis(v, rad);
}

void check_matrix()
{
	using namespace UuIcsGeom;
	Matrix3d m1(1,2,3,0,0,0,0,0,0), m2(0,0,0,4,5,6,0,0,0),
             m3(0,0,0,0,0,0,7,8,9), mr(1,-1,-2,3,-1,-4,2,2,-1);
	Vector3d v1(1,0,0), v2(0,1,0), v3(0,0,1), vr(-2,-4,5);
	Vector3d v = m1*v1;
	v = m1*v2;
	v = m1*v3;
	Matrix3d m = m1*m2;
	m = m1*m3;
	m = m2*m3;
	v=m1*(mr*vr);
	Vector3d vx = (m1*mr)*vr;

	m = rotation_matrix_from_unit_axis(v3, M_PI_2);
	v = m*v1;
	v = m*v2;

	Vector3d axis1(1,1,1);
	m = rotation_matrix_from_axis(axis1, 2*M_PI/3);
	v = m*v1;
	v = m*v2;
	v = m*v3;
	Vector3d axis2(1,2,-3);
	m = rotation_matrix_from_axis(axis2, 1.9);
	v = m*axis2;
	v = m*axis1;

}


void check_basis()
{
	using namespace UuIcsGeom;
	Point3d origin(0.0, 0.0, 0.0), a1(1.0, 0,0), a2(0,0.5,0),
            a2x(0,2,0), a3(0,0,1), c1(1,1,0);
	Basis3d b1(make_basis1(origin, a1, a2));
	Basis3d b2(make_basis1(origin, a1, a2x));
	Basis3d b3(make_basis1(origin, a2, a2x));
	Basis3d b4(make_basis1(origin, a2x, a2x));
	Basis3d b5(make_basis1(origin, c1, a2));
	Vector3d relcoords;
	relcoords = b1.relative_to_this(a3);
	relcoords = b2.relative_to_this(a3);
	relcoords = b5.relative_to_this(a3);
}

Geom::Basis3d Geom::make_basis1(double &def_area, FTriple const &pt1,
				FTriple const &pt2, FTriple const &pt3)
{
	DTriple dpt1, dpt2, dpt3;
	for (int i=0; i!=3; ++i) {
		dpt1[i] = static_cast<double>(pt1[i]);
		dpt2[i] = static_cast<double>(pt2[i]);
		dpt3[i] = static_cast<double>(pt3[i]);
	}
	return make_basis1(def_area, dpt1, dpt2, dpt3);
}

Geom::Basis3d Geom::make_basis1(double &definition_area, DTriple const &pt1,
        DTriple const &pt2, DTriple const &pt3)
{
	Point3d m_reference;
	Vector3d m_axis[3];
	// reference point is barycentre of three points
	for (int i=0; i!=3; ++i) {
		m_reference.coords[i] = (pt1[i]+pt2[i]+pt3[i])/3.0;
	}
	// compute the direction vectors based on two difference vectors
	Vector3d vec[3];
	vec[0] = diffvec(pt1, pt2);
	vec[1] = diffvec(pt1, pt3);
	// The third vector is the cross product of the first two
	vec[2] = cross(vec[0].coords, vec[1].coords);
	double length[3];
	for (int i=0; i!=3; ++i)
		length[i] = vector_length(vec[i].coords);
	definition_area = length[2];
	if (definition_area == 0.0) {
		// degenerate case
		if (length[0]>=length[1]) {
			if (length[0]==0.0) {
				m_axis[0].set(1.0, 0.0, 0.0);
				m_axis[1].set(0.0, 1.0, 0.0);
				m_axis[2].set(0.0, 0.0, 1.0);
				length[0] = length[1] = length[2] = 1.0;
			} else {
				m_axis[0] = vec[0];
				m_axis[1] = a_perpendicular(vec[0].coords);
				length[1] = vector_length(m_axis[1].coords);
				m_axis[2] = cross(m_axis[0].coords, m_axis[1].coords);
				length[2] = vector_length(m_axis[2].coords);
			}
		} else {
				m_axis[1] = vec[1];
				m_axis[0] = a_perpendicular(vec[1].coords);
				length[0] = vector_length(m_axis[0].coords);
				m_axis[2] = cross(m_axis[0].coords, m_axis[1].coords);
				length[2] = vector_length(m_axis[2].coords);
		}
	} else {
		// normal case
		// if (length[0]>=length[1]) {  // don't do this: we want a fixed base system relative to points with an identity
				m_axis[0] = vec[0];
				m_axis[1] = cross(vec[2].coords, vec[0].coords);
				length[1] = vector_length(m_axis[1].coords);
				m_axis[2] = vec[2];
		//} else {
		//		m_axis[0] = cross(vec[1].coords, vec[2].coords);
		//		length[0] = vector_length(m_axis[0].coords);
		//		m_axis[1] = vec[1];
		//		m_axis[2] = vec[2];
		//}
	}
	for (int i=0; i!=3; ++i)
		mult_by(m_axis[i].coords, 1.0/length[i]);
	return Basis3d(m_reference, m_axis[0], m_axis[1], m_axis[2]);
}


Geom::Vector3d Geom::Basis3d::relative_to_this(DTriple const &pt)
{
	DTriple vec;
	for (int i=0; i!=3; ++i) {
		vec[i] = pt[i]-m_reference.coords[i];
	}
	Vector3d result;
	for (int i=0; i!=3; ++i) {
		result.coords[i]=0.0;
		Vector3d const &axis=m_axis[i];
		for (int j=0; j!=3; ++j) {
			result.coords[i] += vec[j]*axis.coords[j];
		}
	}
	return result;
}

